#include <xsTypes.h>
#include <functionMap.h>
#include <stlToCArrays.h>
#include <XSUtil/FunctionUtils/FunctionUtility.h>
#include <XSUtil/Utils/XSutility.h>
#include <cmath>

// Called from C++ to F
extern "C" void xsnteeasb_(float* ear, int& ne, float* pars, int& specnum,
                   float* photar, float* xtot, float* sptot, float* spref);
extern "C" float totlxse_(float* ab, int& In, float* X, int& N, int& fl1,
                   int& fl2, float& Xnor, float* a, float* toth);                   
extern "C" float totlxsep_(float* ab, int& In, float* X, int& N, int& fl1,
                   int& fl2, float& Xnor, float* a, float* toth);                   
 

// Called from F to C++
extern "C" float sigmabfe_(int& In, float* X, int& N, float& Ab0, float& Xnor);
extern "C" float sigbfepr_(int& In, float* X, int& N, float& Ab0, float& Xnor);

void xsnteea(const RealArray& energyArray, const RealArray& params,
        int spectrumNumber, RealArray& flux, RealArray& fluxErr, 
        const string& initString)
{
   using namespace XSutility;

   // C++ entry wrapper to perform simple memory allocation prior to 
   // entering actual nteea model subroutines contained in xsnteeasb.f 
   // (formerly xsnteea.f).
   
   const size_t nE1 = energyArray.size();
   int nE = static_cast<int>(nE1)-1;
   auto_array_ptr<float> apXtot(new float[nE1]);
   auto_array_ptr<float> apSptot(new float[nE1]);
   auto_array_ptr<float> apSpref(new float[nE1]);
   
   float *ear=0, *pars=0, *photar=0, *photer=0;
   XSFunctions::stlToFloatArrays<float>(energyArray, params, flux, fluxErr,
        ear, pars, photar, photer);
   auto_array_ptr<float> apEar(ear);
   auto_array_ptr<float> apPars(pars);
   auto_array_ptr<float> apPhotar(photar);
   auto_array_ptr<float> apPhoter(photer);   

   xsnteeasb_(ear, nE, pars, spectrumNumber, photar, apXtot.get(),
                apSptot.get(), apSpref.get()); 
                
   XSFunctions::floatFluxToStl<float>(photar, photer, nE, false, flux, fluxErr);
    
}

// Routines translated from the Fortran originally in xsnteea.f

// **==sigmabfe.spg  processed by SPAG 4.50J  at 11:13 on 27 Feb 1996
// Translated Jan 2009

float sigmabfe_(int& In, float* X, int& N, float& Ab0, float& Xnor)
{
   // Since this is intended to be called from Fortran, ASSUME In IS
   // 1-BASED.
   
   // bound-free cross section for neutral matter with cosmic abundances
   // per hydrogen atom in units of the Thomson cross section
   // ab0 - iron log10 abundance relative to hydrogen
   // x - energy vector of the length n; simgabf is calculated for
   // element in from x

   static string fgsolr0;
   static bool firstflag = true;
   static int fl1 = 1;
   static int fl2 = 1;
   static float ab[17];
   static float ab_in=0.0;
   static int nold = 0;
   static XSutility::auto_array_ptr<float> apOpac(0);
   static XSutility::auto_array_ptr<float> apXold(0);
   static XSutility::auto_array_ptr<float> apA(0);
   static XSutility::auto_array_ptr<float> apToth(0);
   
   string curAbund(FunctionUtility::ABUND());
   if (curAbund != fgsolr0 || firstflag)
   {
      firstflag = false;
      fl1 = 1;
      fgsolr0 = curAbund;
      ab[0]=12+static_cast<float>(log10(FunctionUtility::getAbundance("H")));
      ab[1]=12+static_cast<float>(log10(FunctionUtility::getAbundance("He")));
      ab[2]=12+static_cast<float>(log10(FunctionUtility::getAbundance("C")));
      ab[3]=12+static_cast<float>(log10(FunctionUtility::getAbundance("N")));
      ab[4]=12+static_cast<float>(log10(FunctionUtility::getAbundance("O")));
      ab[5]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ne")));
      ab[6]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Na")));
      ab[7]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Mg")));
      ab[8]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Al")));
      ab[9]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Si")));
      ab[10]=12+static_cast<float>(log10(FunctionUtility::getAbundance("S")));
      ab[11]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Cl")));
      ab[12]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ar")));
      ab[13]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ca")));
      ab[14]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Cr")));
      ab[15]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Fe")));
      ab[16]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ni")));
      ab_in=ab[15];      
   }
   
   float abh = Ab0 + ab_in;
   // If the energy vector changed from the
   // previous call, recompute the opacity (in totlxs) and reset xold.
   // Remember, In is 1-based.
   if (!apXold.get() || (X[In-1] != apXold[In-1]) || N != nold)
   {
      fl2 = 1;
      float* oldF = apOpac.reset(new float[N]);
      delete [] oldF;
      oldF = apXold.reset(new float[N]);
      delete [] oldF;
      oldF = apA.reset(new float[N*17]);
      delete [] oldF;
      oldF = apToth.reset(new float[N]);
      delete [] oldF;
      nold = N;
      for (int i=0; i<N; ++i)
         apXold[i] = X[i];
   } 
   // If either the iron abundance or the energy vector changed,
   // recompute the opacity
   if (ab[15] != abh)
   {
      fl1 = 1;
      ab[15] = abh;
   }
   if (fl1 || fl2)
   {
      // totlxse expects a 1-based index
      for (int i=1; i<=N; ++i)
         apOpac[i-1] = totlxse_(ab, i, X, N, fl1, fl2, Xnor, apA.get(),
                                apToth.get());
   }
   
   return apOpac[In-1];
} 


//**==sigbfepr.spg  processed by SPAG 4.50J  at 11:13 on 27 Feb 1996
// Translated Jan 2009 
float sigbfepr_(int& In, float* X, int& N, float& Ab0, float& Xnor)
{
   // Since this is intended to be called from Fortran, ASSUME In IS
   // 1-BASED.
   
   //  bound-free cross section for neutral matter with cosmic abundances
   //  per hydrogen atom in units of the Thomson cross section
   //  ab0 - iron log10 abundance relative to hydrogen
   //  x - energy vector of the length n; sigbfepr is calculated for
   //  element in from x
   
   static string fgsolr0;
   static bool firstflag = true;
   static int fl1 = 1;
   static int fl2 = 1;
   static float ab[17];
   static float ab_in=0.0;
   static int nold = 0;
   static XSutility::auto_array_ptr<float> apOpac(0);
   static XSutility::auto_array_ptr<float> apXold(0);
   static XSutility::auto_array_ptr<float> apA(0);
   static XSutility::auto_array_ptr<float> apToth(0);
   
   string curAbund(FunctionUtility::ABUND());
   if (curAbund != fgsolr0 || firstflag)
   {
      firstflag = false;
      fl1 = 1;
      fgsolr0 = curAbund;
      ab[0]=12+static_cast<float>(log10(FunctionUtility::getAbundance("H")));
      ab[1]=12+static_cast<float>(log10(FunctionUtility::getAbundance("He")));
      ab[2]=12+static_cast<float>(log10(FunctionUtility::getAbundance("C")));
      ab[3]=12+static_cast<float>(log10(FunctionUtility::getAbundance("N")));
      ab[4]=12+static_cast<float>(log10(FunctionUtility::getAbundance("O")));
      ab[5]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ne")));
      ab[6]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Na")));
      ab[7]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Mg")));
      ab[8]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Al")));
      ab[9]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Si")));
      ab[10]=12+static_cast<float>(log10(FunctionUtility::getAbundance("S")));
      ab[11]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Cl")));
      ab[12]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ar")));
      ab[13]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ca")));
      ab[14]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Cr")));
      ab[15]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Fe")));
      ab[16]=12+static_cast<float>(log10(FunctionUtility::getAbundance("Ni")));
      ab_in=ab[15];      
   }
   
   float abh = Ab0 + ab_in;
   // If the energy vector changed from the
   // previous call, recompute the opacity (in totlxs) and reset xold.
   // Remember, In is 1-based.
   if (!apXold.get() || (X[In-1] != apXold[In-1]) || N != nold)
   {
      fl2 = 1;
      float* oldF = apOpac.reset(new float[N]);
      delete [] oldF;
      oldF = apXold.reset(new float[N]);
      delete [] oldF;
      oldF = apA.reset(new float[N*17]);
      delete [] oldF;
      oldF = apToth.reset(new float[N]);
      delete [] oldF;
      nold = N;
      for (int i=0; i<N; ++i)
         apXold[i] = X[i];
   } 
   // If either the iron abundance or the energy vector changed,
   // recompute the opacity
   if (ab[15] != abh)
   {
      fl1 = 1;
      ab[15] = abh;
   }
   if (fl1 || fl2)
   {
      // totlxsep expects a 1-based index
      for (int i=1; i<=N; ++i)
         apOpac[i-1] = totlxsep_(ab, i, X, N, fl1, fl2, Xnor, apA.get(),
                                apToth.get());
   }

   return apOpac[In-1];
   
}
